/*
 * Adobe Camera Raw Calibrator
 * Version: 1.1 (x110)
 * http://chromoholics.com/
 *
 * Copyright (c) 2007 Thomas Fors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// =======================================================
// Configuration
// =======================================================

var scriptVersion = "1.1 (x110)";



// =======================================================
// Interface to Adobe Camera Raw
// =======================================================


// -------------------------------------------------------
// Acr()
// Constructor
// -------------------------------------------------------
function Acr() {
	this.temp = 5000;    		// Color Temp: 2000 to 50000
	this.tint = 0;              // Tint: -150 to 150
	this.exps = 0;              // Exposure: -4.0 to +4.0
	this.shad = 5;              // Shadows: 0 to 100
	this.brgh = 50;             // Brightness: 0 to 100
	this.cntr = 25;             // Contrast: -50 to 100
	this.strt = 0;              // Saturation: -100 to 100
	this.shrp = 0;              // Sharpness: 0 to 100
	this.lnr = 0;               // Luminance Smoothing: 0 to 100
	this.cnr = 25;              // Color Noise Reduction: 0 to 100
	this.ca1 = 0;               // Chromatic Aberration (R/C): -100 to 100
	this.ca2 = 0;               // Chromatic Aberration (B/Y): -100 to 100
	this.viga = 0;              // Vignetting Amount: -100 to 100
	this.vigm = 50;             // Vignetting Midpoint: 0 to 100
	this.blkb = 0;              // Shadow Tint: -100 to 100
	this.rhue = 0;              // Red Hue: -100 to 100
	this.rsat = 0;              // Red Saturation: -100 to 100
	this.ghue = 0;              // Green Hue: -100 to 100
	this.gsat = 0;              // Green Saturation: -100 to 100
	this.bhue = 0;              // Blue Hue: -100 to 100
	this.bsat = 0;              // Blue Saturation: -100 to 100
	this.clrs = "ProP";         // Color Space
	this.btdp = "BD8 ";         // Bit Depth
	this.sz = 1024;            	// Image Size (width)  -- ACR will chose closest available
	this.rslt = 240;            // Resolution
	this.reun = "PpIn";         // Resolution Units
	
	this.nOpens = 0;			// Raw file open counter
	
	this.nOpenCalls = 0;
	this.nCloseCalls = 0;
	
	this.version = "Unknown";	// ACR Version
	this.supported = "";
}

// -------------------------------------------------------
// Initialize()
// -------------------------------------------------------
Acr.prototype['Initialize'] = function () {
	this.SetFileName(this.GetActiveDocFileName());
	this.SetCameraModel(this.GetXmpMetadata("tiff:Model"));
	this.SetSize(this.GetActiveDocWidth());
	this.SetColorTemp(this.GetXmpMetadata("crs:Temperature"));
	this.SetTint(this.GetXmpMetadata("crs:Tint"));

	// Fail if this is not the original raw file
	if (activeDocument.name != this.leftTrim(this.GetXmpMetadata("crs:RawFileName"))) {
		throw("You must start AcrCalibrator with the original raw file.\n" +
			  "Do not save the file as a psd or any other file type before\n" +
			  "running the script.");
	};

	// Fail if the raw was cropped or rotated
	if (this.leftTrim(this.GetXmpMetadata("crs:HasCrop")) == "True") {
		throw("Do not crop or rotate the raw file in ACR before running the script.");
	};
	
	this.version = this.GetXmpMetadata("crs:Version");
	this.version = this.version.replace(/[^0123456789\.x]/g, "");
	
	var supportedVersions = Array( "2.2",	 "2.4x101",	"2.4",		"3.0",
								   "3.1",	 "3.2",		"3.3x43",	"3.3",
								   "3.4x89", "3.5",		"3.6",		"3.7",
								   "4.0",    "4.1"  , "4.2" , "4.3" , "4.3.1" );
	
	var supported = false;
	for (var i = 0; i < supportedVersions.length; i++) {
		if (supportedVersions[i] == this.version) {
			supported = true;
		} 
    }        
	
	if ( !supported ) {
		this.supported = " UNSUPPORTED";
		if ( ! confirm("Unsupported Version\n\n" + 
					 "You are running ACR version " + this.version + ".\n" +
					 "This script has not been tested with your version of ACR.\n\n" +
					 "Click:\n" +
					 "          YES to proceed anyway (your results may vary).\n" +
					 "          NO to exit.\n\n" +
					 "Visit http://chromoholics.com and check for script updates.\n\n" +
					 "Do you want to continue?") ) {
			throw("Script aborted at user request.");
		}
	}
}

// -------------------------------------------------------
// GetXmpMetadata()
// Returns the meta data labeled by tag
// -------------------------------------------------------
Acr.prototype['GetXmpMetadata'] = function (tag) {
	var i, metadata, result;
	
	var reFind  = new RegExp("<" + tag + ">.*</" + tag + ">");
	var reStart = new RegExp("<" + tag + ">");
	var reEnd   = new RegExp("</" + tag + ">");
	
	metadata = activeDocument.xmpMetadata.rawData.split("\n");
	result = "";
	for ( i=0; i<metadata.length; i++ ) {
		if ( metadata[i].match(reFind) ) {
			metadata[i] = metadata[i].replace(reStart, "");
			metadata[i] = metadata[i].replace(reEnd, "");
			result = metadata[i];
		}
	}
	
	return result;
}

// -------------------------------------------------------
// leftTrim()
// Utility function to trim whitespace from left of string
// -------------------------------------------------------
Acr.prototype['leftTrim'] = function (str) {
	while (str.substring(0,1) == ' ') {
		str = str.substring(1, str.length);
	}
	return str;
}

// -------------------------------------------------------
// GetActiveDocFileName()
// Returns the Filename of the active document.
// -------------------------------------------------------
Acr.prototype['GetActiveDocFileName'] = function () {
	return activeDocument.fullName;
}

// -------------------------------------------------------
// GetActiveDocWidth()
// Returns the width of the active document.
// -------------------------------------------------------
Acr.prototype['GetActiveDocWidth'] = function () {
	return activeDocument.width;
}

// -------------------------------------------------------
// SetCameraModel()
// -------------------------------------------------------
Acr.prototype['SetCameraModel'] = function (cmod) {
	this.cmod = cmod;
}

// -------------------------------------------------------
// SetFileName()
// -------------------------------------------------------
Acr.prototype['SetFileName'] = function (fname) {
	this.fname = fname;
}

// -------------------------------------------------------
// SetSize()
// -------------------------------------------------------
Acr.prototype['SetSize'] = function (w) {
	this.sz = w;
}

// -------------------------------------------------------
// SetColorTemp()
// -------------------------------------------------------
Acr.prototype['SetColorTemp'] = function (temp) {
	this.temp = parseInt(temp+"");
	if ( isNaN(this.temp) ) {
		this.temp = 5000;
	}
}

// -------------------------------------------------------
// SetTint()
// -------------------------------------------------------
Acr.prototype['SetTint'] = function (tint) {
	this.tint = parseInt(tint+"");
	if ( isNaN(this.tint) ) {
		this.tint = 0;
	}
}

// -------------------------------------------------------
// SetExposure()
// -------------------------------------------------------
Acr.prototype['SetExposure'] = function (exps) {
	this.exps = parseFloat(exps+"");
	if ( isNaN(this.exps) ) {
		this.exps = 0;
	}
}

// -------------------------------------------------------
// SetShadow()
// -------------------------------------------------------
Acr.prototype['SetShadows'] = function (shad) {
	this.shad = parseInt(shad+"");
	if ( isNaN(this.shad) ) {
		this.shad = 0;
	}
}

// -------------------------------------------------------
// SetBrightness()
// -------------------------------------------------------
Acr.prototype['SetBrightness'] = function (brgh) {
	this.brgh = parseInt(brgh+"");
	if ( isNaN(this.brgh) ) {
		this.brgh = 50;
	}
}

// -------------------------------------------------------
// SetContrast()
// -------------------------------------------------------
Acr.prototype['SetContrast'] = function (cntr) {
	this.cntr = parseInt(cntr+"");
	if ( isNaN(this.cntr) ) {
		this.cntr = 25;
	}
}

// -------------------------------------------------------
// SetSaturation()
// -------------------------------------------------------
Acr.prototype['SetSaturation'] = function (strt) {
	this.strt = parseInt(strt+"");
	if ( isNaN(this.strt) ) {
		this.strt = 0;
	}
}

// -------------------------------------------------------
// SetShadowTint()
// -------------------------------------------------------
Acr.prototype['SetShadowTint'] = function (blkb) {
	this.blkb = parseInt(blkb+"");
	if ( isNaN(this.blkb) ) {
		this.blkb = 0;
	}
}

// -------------------------------------------------------
// SetGreenHue()
// -------------------------------------------------------
Acr.prototype['SetGreenHue'] = function (ghue) {
	this.ghue = parseInt(ghue+"");
	if ( isNaN(this.ghue) ) {
		this.ghue = 0;
	}
}

// -------------------------------------------------------
// SetGreenSat()
// -------------------------------------------------------
Acr.prototype['SetGreenSat'] = function (gsat) {
	this.gsat = parseInt(gsat+"");
	if ( isNaN(this.gsat) ) {
		this.gsat = 0;
	}
}

// -------------------------------------------------------
// SetBlueHue()
// -------------------------------------------------------
Acr.prototype['SetBlueHue'] = function (bhue) {
	this.bhue = parseInt(bhue+"");
	if ( isNaN(this.bhue) ) {
		this.bhue = 0;
	}
}

// -------------------------------------------------------
// SetBlueSat()
// -------------------------------------------------------
Acr.prototype['SetBlueSat'] = function (bsat) {
	this.bsat = parseInt(bsat+"");
	if ( isNaN(this.bsat) ) {
		this.bsat = 0;
	}
}

// -------------------------------------------------------
// SetRedHue()
// -------------------------------------------------------
Acr.prototype['SetRedHue'] = function (rhue) {
	this.rhue = parseInt(rhue+"");
	if ( isNaN(this.rhue) ) {
		this.rhue = 0;
	}
}

// -------------------------------------------------------
// SetRedSat()
// -------------------------------------------------------
Acr.prototype['SetRedSat'] = function (rsat) {
	this.rsat = parseInt(rsat+"");
	if ( isNaN(this.rsat) ) {
		this.rsat = 0;
	}
}

// -------------------------------------------------------
// Open()
// Opens RAW file with ACR
// -------------------------------------------------------
Acr.prototype['Open'] = function () {
	var supportedVersions;
	var supported;
	
	this.nOpenCalls++;
	
	var desc1 = new ActionDescriptor();
    desc1.putPath( charIDToTypeID( "null" ), new File( this.fname ) );
	var desc2 = new ActionDescriptor();
	
	// Camera Model
	desc2.putString( charIDToTypeID( "CMod" ), this.cmod );
	
	// Custom Settings
	desc2.putEnumerated( charIDToTypeID( "Sett" ), charIDToTypeID( "Sett" ), charIDToTypeID( "Cst " ) );
	
	// Adjust Tab:
	
	// Custom White Balance
	desc2.putEnumerated( charIDToTypeID( "WBal" ), charIDToTypeID( "WBal" ), charIDToTypeID( "Cst " ) );
	desc2.putInteger( charIDToTypeID( "Temp" ), this.temp );
	desc2.putInteger( charIDToTypeID( "Tint" ), this.tint );
	
	desc2.putDouble( charIDToTypeID( "Exps" ), this.exps );
	desc2.putInteger( charIDToTypeID( "Shad" ), this.shad );
	desc2.putInteger( charIDToTypeID( "Brgh" ), this.brgh );
	desc2.putInteger( charIDToTypeID( "Cntr" ), this.cntr );
	desc2.putInteger( charIDToTypeID( "Strt" ), this.strt );
	
	// Detail Tab:
	
	desc2.putInteger( charIDToTypeID( "Shrp" ), this.shrp );
	desc2.putInteger( charIDToTypeID( "LNR " ), this.lnr );
	desc2.putInteger( charIDToTypeID( "CNR " ), this.cnr );
	
	// Lens Tab:
	
	desc2.putInteger( charIDToTypeID( "CA1 " ), this.ca1 );
	desc2.putInteger( charIDToTypeID( "CA2 " ), this.ca2 );
	desc2.putInteger( charIDToTypeID( "VigA" ), this.viga );
	desc2.putInteger( charIDToTypeID( "VigM" ), this.vigm );
				
	// Calibrate Tab:
				
	desc2.putInteger( charIDToTypeID( "BlkB" ), this.blkb );
	desc2.putInteger( charIDToTypeID( "RHue" ), this.rhue );
	desc2.putInteger( charIDToTypeID( "RSat" ), this.rsat );
	desc2.putInteger( charIDToTypeID( "GHue" ), this.ghue );
	desc2.putInteger( charIDToTypeID( "GSat" ), this.gsat );
	desc2.putInteger( charIDToTypeID( "BHue" ), this.bhue );
	desc2.putInteger( charIDToTypeID( "BSat" ), this.bsat );

	// Settings introduced in ACR 4.0.  Null these out so they don't affect the calibration
	supportedVersions = Array( "4.0",    "4.1"  , "4.2", "4.3" , "4.3.1" );
	supported = false;
	for (var i = 0; i < supportedVersions.length; i++) {
		if (supportedVersions[i] == this.version) {
			supported = true;
		} 
    }        
	
	if (supported) {
		desc2.putBoolean( charIDToTypeID( "CtoG" ), false );		// Color to Grayscale
		desc2.putInteger( charIDToTypeID( "HRec" ), 0 );			// Highlight Recovery
		desc2.putInteger( charIDToTypeID( "FLit" ), 0 );			// Fill Light
		desc2.putInteger( charIDToTypeID( "Vibr" ), 0 );			// Vibrance
		desc2.putInteger( charIDToTypeID( "HA_R" ), 0 );			// Hue Adjust Reds
		desc2.putInteger( charIDToTypeID( "HA_O" ), 0 );			// Hue Adjust Oranges
		desc2.putInteger( charIDToTypeID( "HA_Y" ), 0 );			// Hue Adjust Yellows
		desc2.putInteger( charIDToTypeID( "HA_G" ), 0 );			// Hue Adjust Greens
		desc2.putInteger( charIDToTypeID( "HA_A" ), 0 );			// Hue Adjust Aquas
		desc2.putInteger( charIDToTypeID( "HA_B" ), 0 );			// Hue Adjust Blues
		desc2.putInteger( charIDToTypeID( "HA_P" ), 0 );			// Hue Adjust Purples
		desc2.putInteger( charIDToTypeID( "HA_M" ), 0 );			// Hue Adjust Magentas
		desc2.putInteger( charIDToTypeID( "SA_R" ), 0 );			// Saturation Adjust Reds
		desc2.putInteger( charIDToTypeID( "SA_O" ), 0 );			// Saturation Adjust Oranges
		desc2.putInteger( charIDToTypeID( "SA_Y" ), 0 );			// Saturation Adjust Yellows
		desc2.putInteger( charIDToTypeID( "SA_G" ), 0 );			// Saturation Adjust Greens
		desc2.putInteger( charIDToTypeID( "SA_A" ), 0 );			// Saturation Adjust Aquas
		desc2.putInteger( charIDToTypeID( "SA_B" ), 0 );			// Saturation Adjust Blues
		desc2.putInteger( charIDToTypeID( "SA_P" ), 0 );			// Saturation Adjust Purples
		desc2.putInteger( charIDToTypeID( "SA_M" ), 0 );			// Saturation Adjust Magentas
		desc2.putInteger( charIDToTypeID( "LA_R" ), 0 );			// Luminance Adjust Reds
		desc2.putInteger( charIDToTypeID( "LA_O" ), 0 );			// Luminance Adjust Oranges
		desc2.putInteger( charIDToTypeID( "LA_Y" ), 0 );			// Luminance Adjust Yellows
		desc2.putInteger( charIDToTypeID( "LA_G" ), 0 );			// Luminance Adjust Greens
		desc2.putInteger( charIDToTypeID( "LA_A" ), 0 );			// Luminance Adjust Aquas
		desc2.putInteger( charIDToTypeID( "LA_B" ), 0 );			// Luminance Adjust Blues
		desc2.putInteger( charIDToTypeID( "LA_P" ), 0 );			// Luminance Adjust Purples
		desc2.putInteger( charIDToTypeID( "LA_M" ), 0 );			// Luminance Adjust Magentas
		desc2.putInteger( charIDToTypeID( "STHH" ), 0 );			// Split Toning Highlights Hue
		desc2.putInteger( charIDToTypeID( "STHS" ), 0 );			// Split Toning Highlights Saturation
		desc2.putInteger( charIDToTypeID( "STB " ), 0 );			// Split Toning Balance
		desc2.putInteger( charIDToTypeID( "STSH" ), 0 );			// Split Toning Shadows Hue
		desc2.putInteger( charIDToTypeID( "STSS" ), 0 );			// Split Toning Shadows Saturation
		desc2.putInteger( charIDToTypeID( "PC_S" ), 0 );			// Parametric Curve Shadows
		desc2.putInteger( charIDToTypeID( "PC_D" ), 0 );			// Parametric Curve Darks
		desc2.putInteger( charIDToTypeID( "PC_L" ), 0 );			// Parametric Curve Lights
		desc2.putInteger( charIDToTypeID( "PC_H" ), 0 );			// Parametric Curve Highlights
		desc2.putInteger( charIDToTypeID( "PC_1" ), 25 );			// Parametric Curve Point 1 Position
		desc2.putInteger( charIDToTypeID( "PC_2" ), 50 );			// Parametric Curve Point 2 Position
		desc2.putInteger( charIDToTypeID( "PC_3" ), 75 );			// Parametric Curve Point 3 Position
		var list1 = new ActionList();
		list1.putInteger( 0 );
		list1.putInteger( 0 );
		list1.putInteger( 255 );
		list1.putInteger( 255 );
        desc2.putList( charIDToTypeID( "Crv " ), list1 );			// Linear Tone Curve
	}
	
	// Settings introduced in ACR 4.1.  Null these out so they don't affect the calibration
	supportedVersions = Array( "4.1" , "4.2", "4.3", "4.3.1"  );
	supported = false;
	for (var i = 0; i < supportedVersions.length; i++) {
		if (supportedVersions[i] == this.version) {
			supported = true;
		} 
    }
	if (supported) {
        desc2.putInteger( charIDToTypeID( "Clar" ), 0 );			// Clarity
        desc2.putDouble( charIDToTypeID( "ShpR" ), 1.000000 );		// Sharpening Radius
        desc2.putInteger( charIDToTypeID( "ShpD" ), 25 );			// Sharpening Detail
        desc2.putInteger( charIDToTypeID( "ShpM" ), 0 );			// Sharpening Masking
        desc2.putInteger( charIDToTypeID( "DeFr" ), 0 );			// Lens Corrections Defringe
	}
				
	// Colorspace: ProPhoto RGB
	desc2.putEnumerated( charIDToTypeID( "ClrS" ), charIDToTypeID( "ClrS" ), charIDToTypeID( this.clrs ) );
	// Bit Depth: 8 bits/channel
	desc2.putEnumerated( charIDToTypeID( "BtDp" ), charIDToTypeID( "BtDp" ), charIDToTypeID( this.btdp ) );
	// Size: nearest to 1024 pixels wide
	desc2.putInteger( charIDToTypeID( "Sz  " ), this.sz );
	// Resolution: 240
	desc2.putDouble( charIDToTypeID( "Rslt" ), this.rslt );
	// Resolution Unit: Pixels per Inch
	desc2.putEnumerated( charIDToTypeID( "ReUn" ), charIDToTypeID( "ReUn" ), charIDToTypeID( this.reun ) );
				
	desc1.putObject( charIDToTypeID( "As  " ), stringIDToTypeID( "Adobe Camera Raw" ), desc2 );

	try {
		executeAction( charIDToTypeID( "Opn " ), desc1, DialogModes.NO );
	} catch (e) {
		throw("Incorrect or missing file:\n\n" +
			  "The file:\n     " + this.fname + "\n" +
			  "is either missing or not a camera raw file.\n\n" +
			  "Please open a camera raw file before running the script.");
	}
				
	cc.MarkAllPatchesUnread();
				
	// Increment raw file open counter
	this.nOpens++;
}

// -------------------------------------------------------
// Close()
// Closes active document without saving
// -------------------------------------------------------
Acr.prototype['Close'] = function () {
	this.nCloseCalls++;
	activeDocument.close(SaveOptions.DONOTSAVECHANGES);
}


// =======================================================
// GretagMacbeth Color Checker Routines
// =======================================================


// -------------------------------------------------------
// ColorChecker()
// Constructor
// -------------------------------------------------------
function ColorChecker() {
	
	this.x = new Array();						// array of x-axis pixel coordinates of patches
	this.y = new Array();						// array of y-axiz pixel coordinates of patches
	this.refColor = new Array();				// array of reference patch colors
	this.measColor = new Array();				// array of measured patch colors
	this.bRead = new Array();					// array of boolean flags indicating patch has been read
	
	// Pro Photo RGB Reference
	this.SetReferenceRGB( 0,  82,  66,  50);	// dark skin
	this.SetReferenceRGB( 1, 162, 140, 119);	// light skin
	this.SetReferenceRGB( 2,  96, 104, 134);	// blue sky
	this.SetReferenceRGB( 3,  77,  89,  55);	// foliage
	this.SetReferenceRGB( 4, 122, 114, 157);	// blue flower
	this.SetReferenceRGB( 5, 131, 168, 158);	// bluish green
	this.SetReferenceRGB( 6, 163, 118,  43);	// orange
	this.SetReferenceRGB( 7,  78,  75, 138);	// purplish blue
	this.SetReferenceRGB( 8, 144,  84,  78);	// moderate red
	this.SetReferenceRGB( 9,  69,  52,  83);	// purple
	this.SetReferenceRGB(10, 145, 168,  73);	// yellow green
	this.SetReferenceRGB(11, 182, 155,  60);	// orange yellow
	this.SetReferenceRGB(12,  59,  49, 127);	// blue
	this.SetReferenceRGB(13,  85, 124,  65);	// green
	this.SetReferenceRGB(14, 122,  56,  43);	// red
	this.SetReferenceRGB(15, 197, 186,  65);	// yellow
	this.SetReferenceRGB(16, 144,  85, 124);	// magenta
	this.SetReferenceRGB(17,  75, 109, 144);	// cyan
	this.SetReferenceRGB(18, 237, 237, 234);	// white
	this.SetReferenceRGB(19, 187, 190, 187);	// neutral 8
	this.SetReferenceRGB(20, 145, 145, 143);	// neutral 6.5
	this.SetReferenceRGB(21, 104, 104, 103);	// neutral 5
	this.SetReferenceRGB(22,  67,  67,  66);	// neutral 3.5
	this.SetReferenceRGB(23,  38,  38,  37);	// black
	
	this.MarkAllPatchesUnread();
}

// -------------------------------------------------------
// SetReferenceRGB()
// Sets the reference patch (n) color based on specified
// r, b, g.
// -------------------------------------------------------
ColorChecker.prototype['SetReferenceRGB'] = function (n, r, g, b) {
	this.refColor[n] = new ProPhotoRGBColor();
	this.refColor[n].SetRGB(r, g, b);
}

// -------------------------------------------------------
// MarkAllPatchesUnread()
// -------------------------------------------------------
ColorChecker.prototype['MarkAllPatchesUnread'] = function () {
	var i;
	
	for ( i=0; i<24; i++ ) {
		this.bRead[i] = 0;
	}
}

// -------------------------------------------------------
// FindPatchesFromActiveDoc()
// Reads X,Y coordinates from a four-corner pen path
// in active document and computes color checker patch
// coordinates.
// -------------------------------------------------------
ColorChecker.prototype['FindPatchesFromActiveDoc'] = function () {
	var tlX, tlY, blX, blY, brX, brY, trX, trY;
	
	// Read path to find corner patches
	if ( activeDocument.pathItems.length == 1 ) {	
		// document contains a single path
		if ( activeDocument.pathItems[0].subPathItems[0].pathPoints.length == 4 ) {  
			// path has exactly four points
			tlX = activeDocument.pathItems[0].subPathItems[0].pathPoints[0].anchor[0];
			tlY = activeDocument.pathItems[0].subPathItems[0].pathPoints[0].anchor[1];
			blX = activeDocument.pathItems[0].subPathItems[0].pathPoints[1].anchor[0];
			blY = activeDocument.pathItems[0].subPathItems[0].pathPoints[1].anchor[1];
			brX = activeDocument.pathItems[0].subPathItems[0].pathPoints[2].anchor[0];
			brY = activeDocument.pathItems[0].subPathItems[0].pathPoints[2].anchor[1];
			trX = activeDocument.pathItems[0].subPathItems[0].pathPoints[3].anchor[0];
			trY = activeDocument.pathItems[0].subPathItems[0].pathPoints[3].anchor[1];
		}
		else {
			throw("Path must have exactly four points");
		}
	}
	else {
		throw ("You must define a path");
	}
	
	this.patchCornerString  = "("+Math.round(tlX)+","+Math.round(tlY)+") ";
	this.patchCornerString += "("+Math.round(blX)+","+Math.round(blY)+") ";
	this.patchCornerString += "("+Math.round(brX)+","+Math.round(brY)+") ";
	this.patchCornerString += "("+Math.round(trX)+","+Math.round(trY)+")";
	
	// X,Y coords of corner patches (tl, bl, br, tr)
	this.ComputePatchCoordinates(tlX,tlY, blX,blY, brX,brY, trX,trY);
}

// -------------------------------------------------------
// ComputePatchCoordinates()
// Compute patch locations based on corner X,Y coordinates
// -------------------------------------------------------
ColorChecker.prototype['ComputePatchCoordinates'] = function (tlX,tlY, blX,blY, brX,brY, trX,trY) {
	var row, col;
	
	for ( row = 0; row < 4; row++ ) {
		for ( col = 0; col<6; col++ ) {
			idx = row * 6 + col;  					// patch index
			
			lX = row * (blX - tlX) / 3 + tlX; 		// left x coord of row
			rX = row * (brX - trX) / 3 + trX; 		// right x coord of row
			this.x[idx] = col * (rX - lX) / 5 + lX;	// x coord of patch
			this.x[idx] = Math.round(this.x[idx]);
			
			tY = col * (trY - tlY) / 5 + tlY; 		// top y coord of col
			bY = col * (brY - blY) / 5 + blY; 		// top y coord of col
			this.y[idx] = row * (bY - tY) / 3 + tY;	// y coord of patch
			this.y[idx] = Math.round(this.y[idx]);
		}
	}
	
	// Compute area (K) of quadrilateral abcd with diagonals p,q using Bretschneider's formula
	var a = Math.sqrt(Math.pow(this.x[18]-this.x[ 0], 2) + Math.pow(this.y[18]-this.y[ 0], 2));
	var b = Math.sqrt(Math.pow(this.x[ 0]-this.x[ 5], 2) + Math.pow(this.y[ 0]-this.y[ 5], 2));
	var c = Math.sqrt(Math.pow(this.x[ 5]-this.x[23], 2) + Math.pow(this.y[ 5]-this.y[23], 2));
	var d = Math.sqrt(Math.pow(this.x[23]-this.x[18], 2) + Math.pow(this.y[23]-this.y[18], 2));
	var p = Math.sqrt(Math.pow(this.x[ 0]-this.x[23], 2) + Math.pow(this.y[ 0]-this.y[23], 2));
	var q = Math.sqrt(Math.pow(this.x[18]-this.x[ 5], 2) + Math.pow(this.y[18]-this.y[ 5], 2));
	var K = Math.sqrt(4*p*p*q*q - Math.pow((b*b + d*d - a*a - c*c), 2)) / 4;
	
	// Set max marquee size as fraction of quadrilateral area (K)
	this.maxMarqueeSize = Math.round(Math.sqrt(K/16));
}

// -------------------------------------------------------
// SelectRectangle()
// Select a rectangle defined by its edges
// -------------------------------------------------------
ColorChecker.prototype['SelectRectangle'] = function (t, l, b, r) {
	activeDocument.selection.select(	Array(Array(Math.round(l), Math.round(t)),
											  Array(Math.round(r), Math.round(t)), Array(Math.round(r), Math.round(b)),
											  Array(Math.round(l), Math.round(b)), Array(Math.round(l), Math.round(t))) );
}

// -------------------------------------------------------
// PrevHistState()
// Restore previous history state
// -------------------------------------------------------
ColorChecker.prototype['PrevHistState'] = function () {
    var desc1 = new ActionDescriptor();
	var ref1 = new ActionReference();
	ref1.putEnumerated( charIDToTypeID( "HstS" ), charIDToTypeID( "Ordn" ), charIDToTypeID( "Prvs" ) );
    desc1.putReference( charIDToTypeID( "null" ), ref1 );
	
	executeAction( charIDToTypeID( "slct" ), desc1, DialogModes.NO );
}

// -------------------------------------------------------
// GetChannelHist()
// Get histogram for a particular channel: "Rd  ", "Grn ", or "Bl  "
// -------------------------------------------------------
ColorChecker.prototype['GetChannelHist'] = function (chan) {
    var desc1 = new ActionDescriptor();
    desc1.putBoolean( charIDToTypeID( "Mnch" ), true );
	var desc2 = new ActionDescriptor();
	desc2.putUnitDouble( charIDToTypeID( chan ), charIDToTypeID( "#Prc" ), 100.000000 );
    desc1.putObject( charIDToTypeID( "Gry " ), charIDToTypeID( "ChMx" ), desc2 );
	
	executeAction( charIDToTypeID( "ChnM" ), desc1, DialogModes.NO );
	
    var rv = activeDocument.histogram;
	
    this.PrevHistState();
	
    return rv;
}

// -------------------------------------------------------
// ReadPatch()
// Read a single ColorChecker patch
// -------------------------------------------------------
ColorChecker.prototype['ReadPatch'] = function (idx) {
	var color;
	var nTries = 0;
	var bSuccess = false;
	
	while ( !bSuccess && nTries < 3 ) {
		try {
			if ( this.bRead[idx] != 0 ) {
				color = this.measColor[idx];
			} else {
				var i;
				color = new SolidColor();
				var size = this.maxMarqueeSize/12;
				this.SelectRectangle(this.y[idx]-size, this.x[idx]-size, this.y[idx]+size, this.x[idx]+size);
				
				// Get data
				var histRed = this.GetChannelHist("Rd  ");
				var histGrn = this.GetChannelHist("Grn ");
				var histBlu = this.GetChannelHist("Bl  ");

				this.PrevHistState();
				
				// Get total pixel count
				var totRed = 0;
				var totGrn = 0;
				var totBlu = 0;
				for ( i=0; i<256; i++ ) {
					totRed += histRed[i];
					totGrn += histGrn[i];
					totBlu += histBlu[i];
				}
				
				// Compute r,g,b
				var red = 0;
				var grn = 0;
				var blu= 0;
				for ( i=0; i<256; i++ ) {
					red += i * histRed[i] / totRed;
					grn += i * histGrn[i] / totGrn;
					blu += i * histBlu[i] / totBlu;
				}
				
				// return color
				color = new ProPhotoRGBColor();
				color.SetRGB(red, grn, blu);
				
				// indicate that patch was read for future lookups
				this.measColor[idx] = color;
				this.bRead[idx] = 1;
			}
			
			bSuccess = true;
		} catch (e) {
			// raw file was not active window.  Select it and try again.
			nTries++;
			display.SelectRawWnd();
		}
	}
	
	return color;
}

// -------------------------------------------------------
// GetWbError()
// Returns white balance error of patch 19
// -------------------------------------------------------
ColorChecker.prototype['GetWbError'] = function () {
	acr.Open();
	var color = this.ReadPatch(19);
	var err = Math.sqrt((color.a*color.a + color.b*color.b));
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetGrayError()
// Returns RMS luminance error of gray patches
// -------------------------------------------------------
ColorChecker.prototype['GetGrayError'] = function () {
	acr.Open();
	var err = 0;
	err += 1 * Math.pow( (cc.refColor[19].L - cc.ReadPatch(19).L), 2 );
	err += 1 * Math.pow( (cc.refColor[20].L - cc.ReadPatch(20).L), 2 );
	err += 1 * Math.pow( (cc.refColor[21].L - cc.ReadPatch(21).L), 2 );
	err += 1 * Math.pow( (cc.refColor[22].L - cc.ReadPatch(22).L), 2 );
	err += 1 * Math.pow( (cc.refColor[23].L - cc.ReadPatch(23).L), 2 );
	err /= 5;
	err = Math.sqrt(err); // RMS
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetSatError()
// Returns saturation error of green patch
// -------------------------------------------------------
ColorChecker.prototype['GetSatError'] = function (settings) {
	var err = 0;
	
	acr.Open();
	err += Math.pow( (cc.refColor[13].G - cc.ReadPatch(13).G), 2 );
	err = Math.sqrt(err);
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetShadowTintError()
// Returns white balance error of patch 23
// -------------------------------------------------------
ColorChecker.prototype['GetShadowTintError'] = function () {
	acr.Open();
	var color = this.ReadPatch(23);
	var err = Math.sqrt((color.a*color.a + color.b*color.b));
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetGreenError()
// Returns error of green patch
// -------------------------------------------------------
ColorChecker.prototype['GetGreenError'] = function () {
	acr.Open();
	var err = Math.sqrt( Math.pow( (cc.refColor[13].a - cc.ReadPatch(13).a), 2 ) + Math.pow( (cc.refColor[13].b - cc.ReadPatch(13).b), 2 ) );
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetBlueError()
// Returns error of blue patch
// -------------------------------------------------------
ColorChecker.prototype['GetBlueError'] = function () {
	acr.Open();
	var err = Math.sqrt( Math.pow( (cc.refColor[12].a - cc.ReadPatch(12).a), 2 ) + Math.pow( (cc.refColor[12].b - cc.ReadPatch(12).b), 2 ) );
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// GetRedError()
// Returns error of red patch
// -------------------------------------------------------
ColorChecker.prototype['GetRedError'] = function () {
	acr.Open();
	var err = Math.sqrt( Math.pow( (cc.refColor[14].a - cc.ReadPatch(14).a), 2 ) + Math.pow( (cc.refColor[14].b - cc.ReadPatch(14).b), 2 ) );
	acr.Close();
	
	return err;
}

// -------------------------------------------------------
// WhiteBalance()
// Performs a white balance
// -------------------------------------------------------
ColorChecker.prototype['WhiteBalance'] = function (n) {
	var settings;
	var optimizer;
	
	var msg = "White Balance";
	if ( n == 2 )
		msg = "White Balance (2nd pass)";
	
	optimizer = new Simplex(2, msg);
	optimizer.SetDimension(0, 2000, 50000, 50, "SetColorTemp");
	optimizer.SetDimension(1, -150, 150, 1, "SetTint");
	optimizer.DefineOptimization(-1, "GetWbError", 0.1, undefined, undefined, 0);
	settings = new Array(2);
	settings[0] = acr.temp;
	settings[1] = acr.tint;
	optimizer.Optimize(settings);	// initial settings
	display.temp = acr.temp;
	display.tint = acr.tint;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// GrayAdjust()
// Optimizes gray patches
// -------------------------------------------------------
ColorChecker.prototype['GrayAdjust'] = function () {
	var settings;
	var optimizer;
	
	optimizer = new Simplex(4, "Gray Patches");
	optimizer.SetDimension(0, -4, 4, .05, "SetExposure");
	optimizer.SetDimension(1, 0, 100, 1, "SetShadows");
	optimizer.SetDimension(2, 0, 150, 1, "SetBrightness");
	optimizer.SetDimension(3, -50, 100, 1, "SetContrast");
	optimizer.DefineOptimization(-1, "GetGrayError", 0, undefined, undefined, 0);
	settings = new Array(4);
	settings[0] = acr.exps;
	settings[1] = acr.shad;
	settings[2] = acr.brgh;
	settings[3] = acr.cntr;
	optimizer.Optimize(settings);	// initial settings
	display.exps = acr.exps;
	display.shad = acr.shad;
	display.brgh = acr.brgh;
	display.cntr = acr.cntr;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// SaturationAdjust()
// Optimizes saturation control for green component of
// green patch
// -------------------------------------------------------
ColorChecker.prototype['SaturationAdjust'] = function () {
	//
	// Disabled since it appears to cause more damage than good
	//
	
	//var settings;
	//var optimizer;
	
	//optimizer = new Simplex(1, "Working on saturation...");
	//optimizer.SetDimension(0, -100, 100, 1, "SetSaturation");
	//optimizer.DefineOptimization(-1, "GetSatError", 0.01, 10, undefined, 1);
	//settings = new Array(1);
	//settings[0] = acr.strt;
	//optimizer.Optimize(settings);	// initial settings
	display.strt = acr.strt;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// ShadowTintAdjust()
// Optimizes shadow tint control for neutrality in
// black patch
// -------------------------------------------------------
ColorChecker.prototype['ShadowTintAdjust'] = function () {
	var settings;
	var optimizer;
	
	optimizer = new Simplex(1, "Shadow Tint");
	optimizer.SetDimension(0, -100, 100, 1, "SetShadowTint");
	optimizer.DefineOptimization(-1, "GetShadowTintError", 0, undefined, undefined, 0);
	settings = new Array(1);
	settings[0] = acr.blkb;
	optimizer.Optimize(settings);	// initial settings
	display.blkb = acr.blkb;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// GreenCalibration()
// Adjusts the green controls on the ACR calibration tab
// -------------------------------------------------------
ColorChecker.prototype['GreenCalibration'] = function (n) {
	var settings;
	var optimizer;
	
	var msg = "Green Calibration";
	if ( n == 2 )
		msg = "Green Calibration (2nd pass)";
	if ( n == 3 )
		msg = "Green Calibration (3rd pass)";
	
	optimizer = new Simplex(2, msg);
	optimizer.SetDimension(0, -100, 100, 1, "SetGreenHue");
	optimizer.SetDimension(1, -100, 100, 1, "SetGreenSat");
	optimizer.DefineOptimization(-1, "GetGreenError", 0, undefined, undefined, 0);
	settings = new Array(2);
	settings[0] = acr.ghue;
	settings[1] = acr.gsat;
	optimizer.Optimize(settings);	// initial settings
	display.ghue = acr.ghue;
	display.gsat = acr.gsat;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// BlueCalibration()
// Adjusts the blue controls on the ACR calibration tab
// -------------------------------------------------------
ColorChecker.prototype['BlueCalibration'] = function (n) {
	var settings;
	var optimizer;
	
	var msg = "Blue Calibration";
	if ( n == 2 )
		msg = "Blue Calibration (2nd pass)";
	if ( n == 3 )
		msg = "Blue Calibration (3rd pass)";
	
	optimizer = new Simplex(2, msg);
	optimizer.SetDimension(0, -100, 100, 1, "SetBlueHue");
	optimizer.SetDimension(1, -100, 100, 1, "SetBlueSat");
	optimizer.DefineOptimization(-1, "GetBlueError", 0, undefined, undefined, 0);
	settings = new Array(2);
	settings[0] = acr.bhue;
	settings[1] = acr.bsat;
	optimizer.Optimize(settings);	// initial settings
	display.bhue = acr.bhue;
	display.bsat = acr.bsat;
	display.DisplayStatus(" ");
}

// -------------------------------------------------------
// RedCalibration()
// Adjusts the red controls on the ACR calibration tab
// -------------------------------------------------------
ColorChecker.prototype['RedCalibration'] = function (n) {
	var settings;
	var optimizer;
	
	var msg = "Red Calibration";
	if ( n == 2 )
		msg = "Red Calibration (2nd pass)";
	if ( n == 3 )
		msg = "Red Calibration (3rd pass)";
	
	optimizer = new Simplex(2, msg);
	optimizer.SetDimension(0, -100, 100, 1, "SetRedHue");
	optimizer.SetDimension(1, -100, 100, 1, "SetRedSat");
	optimizer.DefineOptimization(-1, "GetRedError", 0, undefined, undefined, 0);
	settings = new Array(2);
	settings[0] = acr.rhue;
	settings[1] = acr.rsat;
	optimizer.Optimize(settings);	// initial settings
	display.rhue = acr.rhue;
	display.rsat = acr.rsat;
	display.DisplayStatus(" ");
}


// =======================================================
// Variable-size N-dimensional simplex optimization.
// =======================================================



// -------------------------------------------------------
// Simplex()
// Constructor
// -------------------------------------------------------
function Simplex(nDim, name) {
	var i;
	
	this.name = name;
	this.nDim = nDim;					// number of dimensions
	
	this.uv = new Array();				// array of vertices in user coordinates
	this.v = new Array();				// array of vertices in normalized coordinates
	this.r = new Array();				// array of results
	
	this.rCache = new Array();			// result cache
	this.rCacheSize = 0;				// cache size
	this.rCacheHits = 0;				// number of cache hits
	
	this.bestSettings = new Array();	// best settings found
	
	// Construct vertex arrays
	for ( i=0; i<this.nDim+1+4; i++ ) {
		this.uv[i] = new Array(this.nDim);
		this.v[i] = new Array(this.nDim);
		this.r[i] = new Array(this.nDim);
	}
	
	i = this.nDim+1;
	this.R = i++;				// index of Reflected vertex
	this.E = i++;				// index of Expanded vertex
	this.Cr = i++;				// index of Contracted vertex on Wastebasket side
	this.Cw = i++;				// index of Contracted vertex on Reflected side
	
	this.W = -1;				// index of Wastebasket vertex
	this.N = -1;				// index of Next-worst vertex
	this.B = -1;				// index of Best vertex
	
	this.uLo = new Array();		// low boundary in user-space for each dimension
	this.uHi = new Array();		// high boundary in user-space for each dimension
	this.uRes = new Array();	// resolution required in user-space for each dimension
	this.fnSet = new Array();	// array of functions to set each parameter (dimension)
	
	// Load cache
	if ( display.files["cache"] != undefined ) {
		var cacheIndex, cacheValue;
		display.files["cache"].open("e");
		while ( !display.files["cache"].eof ) {
			cacheIndex = display.files["cache"].readln();
			cacheValue = display.files["cache"].readln();
			if ( cacheIndex.length > 0 ) {
				this.rCache[cacheIndex] = cacheValue;
				this.rCacheSize++;
			}
		}
		display.files["cache"].close();
	}
}


// -------------------------------------------------------
// SetDimension()
// Sets user-space boundary and resolution requirements
// -------------------------------------------------------
Simplex.prototype['SetDimension'] = function (d, uLo, uHi, resolution, fnSet) {
	this.uLo[d] = uLo;
	this.uHi[d] = uHi;
	this.uRes[d] = resolution;
	this.fnSet[d] = fnSet;
}

// -------------------------------------------------------
// NormalizeVertex()
// Convert vertex from user space to normalized space
// -------------------------------------------------------
Simplex.prototype['NormalizeVertex'] = function (i) {
	var d, m, b;
	for ( d=0; d<this.nDim; d++ ) {
		m = 1/(this.uHi[d]-this.uLo[d]);
		b = m * (0 - this.uLo[d]);
		this.v[i][d] = this.uv[i][d] * m + b;
	}
}

// -------------------------------------------------------
// DenormalizeVertex()
// Convert vertex from normalized space to user space
// -------------------------------------------------------
Simplex.prototype['DenormalizeVertex'] = function (i) {
	var d, m, b;
	for ( d=0; d<this.nDim; d++ ) {
		m = this.uHi[d]-this.uLo[d];
		b = this.uLo[d];
		this.uv[i][d] = Math.round( (this.v[i][d] * m + b) / this.uRes[d] ) * this.uRes[d];
	}
}

// -------------------------------------------------------
// ProjectVertex()
// Project vertex (i) of simplex through average of remaining
// vertices by amount (factor).
// -------------------------------------------------------
Simplex.prototype['ProjectVertex'] = function (i, factor) {
	var d, j, center;
	var newVertex = new Array();
	
	// for each dimension...
	for ( d=0; d<this.nDim; d++ ) {
		// compute center of remaining vertices
		center = -this.v[i][d];
		for ( j=0; j<this.nDim+1; j++ ) {
			center += this.v[j][d];
		}
		center /= this.nDim;	// num of remaining vertices
		
		// compute new vertex
		newVertex[d] = (1-factor)*center + (factor)*this.v[i][d];
	}
	
	return newVertex;
}

// -------------------------------------------------------
// ComputeReflections()
// Computes R, E, Cr, and Cw vertices from W
// -------------------------------------------------------
Simplex.prototype['ComputeReflections'] = function () {
	this.v[this.R]  = this.ProjectVertex(this.W, -1);
	this.v[this.E]  = this.ProjectVertex(this.W, -2);
	this.v[this.Cr] = this.ProjectVertex(this.W, -0.5);
	this.v[this.Cw] = this.ProjectVertex(this.W, 0.5);
}

// -------------------------------------------------------
// RankVertices()
// Ranks the vertices assigning W, N, and B
// Larger numbers (more positive) are better
// -------------------------------------------------------
Simplex.prototype['RankVertices'] = function () {
	var i;
	var worst;
	var nextWorst;
	var best;
	
	// Initialize W, N, and B before searching
	best = 0;
	if ( this.r[0] > this.r[1] ) {
		worst = 0;
		nextWorst = 1;
	} else {
		worst = 1;
		nextWorst = 0;
	}
	
	// Search...
	for ( i=0; i<this.nDim+1; i++ ) {
		if ( this.r[i] < this.r[worst] ) {
			nextWorst = worst;		// new Next-worst
			worst = i;				// new Worst
		} else if ( this.r[i] < this.r[nextWorst] && i != worst ) {
			nextWorst = i;			// new Worst-next
		}
		if ( this.r[i] >= this.r[best] ) {
			best = i;				// new Best
		}
	}
	
	// Make the assignments
	if ( this.W < 0 ) {
		// we're ranking vertices of the first simplex, so we assign W
		// to the absolute worst point
		this.W = worst;
		this.N = nextWorst;
	} else {
		// otherwise W will have been assigned to the previous N in
		// Step() so we honor that here and just assign N (and B).
		if ( this.W != worst ) {
			this.N = worst;
		}	else {
			this.N = nextWorst;
		}
	}
	// B is always best
	this.B = best;
	
	// Check for new best result and save
	if ( this.r[this.B] > this.bestResult ) {
		this.bestResult = this.r[this.B];
		this.DenormalizeVertex(this.B);
		for ( d=0; d<this.nDim; d++ ) {
			this.bestSettings[d] = this.uv[this.B][d];
		}
	}
}

// -------------------------------------------------------
// DefineOptimization()
// type = -1 for minmization,
//   0 for zero search (untested),
//  +1 for maximization (untested).
// fnEval must accept an array of nDim control values and return a result
// Optimization will return when best-case result beats goal
//   or timeLimit (in minutes) has been exceeded
//   or stepLimit has been exceeded
//   or restartLimit has been exceeded
// if any of these limits are set to undefined then they are disabled
// -------------------------------------------------------
Simplex.prototype['DefineOptimization'] = function (type, fnEval, goal, timeLimit, stepLimit, restartLimit) {
	this.optType = type;
	this.fnEval = fnEval;
	this.goal = goal;
	this.timeLimit = ( timeLimit == undefined ) ? (Number.POSITIVE_INFINITY) : (timeLimit);
	this.stepLimit = ( stepLimit == undefined ) ? (Number.POSITIVE_INFINITY) : (stepLimit);
	this.restartLimit = ( restartLimit == undefined ) ? (Number.POSITIVE_INFINITY) : (restartLimit);
}

// -------------------------------------------------------
// EvaluateVertex()
// Evaluate vertex and record result.  If vertex is out
// of bounds, return highly undesirable (Number.NEGATIVE_INFINITY)
// score.
// -------------------------------------------------------
Simplex.prototype['EvaluateVertex'] = function (i) {
	var d, result;
	var cacheIndex;
	
	//  Test for out of bounds condition
	var bOutOfBounds = 0;
	for ( d=0; d<this.nDim; d++ ) {
		if ( this.v[i][d] < 0 || this.v[i][d] > 1 ) {
			bOutOfBounds = 1;
		}
	}
	
	this.DenormalizeVertex(i);
	
	var bShowStatus = true;
	if ( bOutOfBounds ) {
		this.r[i] = Number.NEGATIVE_INFINITY;	// return extremely bad result to force a contraction
		bShowStatus = false;
	} else {
		// check cache first.  if not present, do the more time-consuming evaluation
		cacheIndex = "";
		for ( d=0; d<this.nDim; d++ ) {
			cacheIndex += this.uv[i][d] + " ";
		}
		result = this.rCache[cacheIndex];
		if ( result == undefined ) {
			for ( d=0; d<this.nDim; d++ ) {
				acr[this.fnSet[d]](this.uv[i][d]);	// make settings
			}
			result = cc[this.fnEval]();				// evaluate
			this.rCache[cacheIndex] = result;
			this.rCacheSize++;
		} else {
			this.rCacheHits++;
			bShowStatus = false;
		}
		
		// count vertex evaluations for degenerate detection
		try {
			(this.evalCount[cacheIndex] == undefined) ? this.evalCount[cacheIndex]=1 : this.evalCount[cacheIndex]++;
		} catch (e) {
			// ignore exceptions
		}
		
		if ( this.optType < 0 ) {
			this.r[i] = -result;			// minimization: optimize for -result since RankVertices() assumes maximization
		} else if ( this.optType > 0 ) {
			this.r[i] = result;				// maximization: this is the default for RankVertices()
		} else {
			this.r[i] = -Math.abs(result);	// zero: optimize for -abs(result) since RankVertices() assumes maximization
		}
	}
	
	var ch;
	ch = "   ";
	switch (i) {
		case this.B: ch = "B  "; break;
		case this.N: ch = "N  "; break;
		case this.W: ch = "W  "; break;
		case this.R: ch = "R  "; break;
		case this.E: ch = "E  "; break;
		case this.Cr: ch = "Cr "; break;
		case this.Cw: ch = "Cw "; break;
	}
	
	// Display status
	if ( bShowStatus ) {
		var msg = "Working on " + this.name + "...\r";
		msg += "  Best Result: [";
		for ( d=0; d<this.nDim; d++ ) {
			msg += " " + this.bestSettings[d];
		}
		msg += " ]=" + Math.abs(-Math.round(this.bestResult*10000)/10000) + "\r";
		msg += "  Current Attempt: ["
			for ( d=0; d<this.nDim; d++ ) {
				msg += " " + this.uv[i][d];
			}
		msg += " ]=" + Math.abs(Math.round(this.r[i]*10000)/10000) + "\r";
		display.DisplayStatus(msg);
	}
}


// -------------------------------------------------------
// EvaluateResults()
// -------------------------------------------------------
Simplex.prototype['EvaluateResults'] = function (goal) {
	display.PostResults(this.name+":", (Math.abs(this.bestResult)<goal), Math.abs(this.bestResult));
	
	return Math.abs(this.bestResult);
}


// -------------------------------------------------------
// Initialize()
// Constructs an initial tilted simplex based on provided startingPoint
// which must be an array of nDim user-space coordinates.
// -------------------------------------------------------
Simplex.prototype['Initialize'] = function (startingPoint) {
	var d, i;
	var step, p, q;
	
	this.evalCount = new Array();	// vertex evaluation count for degenerate detection
	this.W = -1;					// index of Wastebasket vertex
	this.N = -1;					// index of Next-worst vertex
	this.B = -1;					// index of Best vertex
	
	for ( d=0; d<this.nDim; d++ ) {
		this.uv[0][d] = startingPoint[d];
	}
	this.NormalizeVertex(0);
	
	step = new Array();
	for ( d=0; d<this.nDim; d++ ) {
		step[d] = (this.v[0][d] > 0.75) ? (-0.25) : (0.25);
	}
	p = (Math.sqrt(this.nDim+1) + this.nDim - 1) / (this.nDim * Math.sqrt(2));
	q = (Math.sqrt(this.nDim+1) - 1) / (this.nDim * Math.sqrt(2));
	
	// Build the tilted simplex
	for ( i=1; i<this.nDim+1; i++ ) {
		for ( d=0; d<this.nDim; d++ ) {
			this.v[i][d] = this.v[0][d] + step[d] * ( (i-1 == d) ? (p) : (q) );
		}
	}
	
	// Evaluate the initial simplex
	for ( i=0; i<this.nDim+1; i++ ) {
		this.EvaluateVertex(i);
	}
	
	this.W = -1;		// index of Wastebasket vertex
	this.N = -1;		// index of Next-worst vertex
	this.B = -1;		// index of Best vertex
}

// -------------------------------------------------------
// Step()
// Perform an optimization step.  This may result in a move,
// expansion, or contraction of the simplex.
//
// Note:  If R is out of bounds, Cr is evaluated.  If Cr
// is in bounds, a Cr contraction is done rather than the
// default Cw contraction.  This is more in keeping with
// the spirit of the algorithm.
// -------------------------------------------------------
Simplex.prototype['Step'] = function () {
	var d;
	var useVertex;
	
	// step 1
	this.RankVertices();
	this.ComputeReflections();
	
	// step 2
	this.EvaluateVertex(this.R);
	if ( this.r[this.N] <= this.r[this.R] && this.r[this.R] <= this.r[this.B] ) {
		// We're moving in a mediocre direction.  Just accept it.
		// N <= R <= B    use simplex B..NR and go to step 3
		useVertex = this.R;
	} else if ( this.r[this.R] > this.r[this.B] ) {
		// We're moving in a desirable direction.  Try expanding.
		// R > B    calculate and evaluate E
		this.EvaluateVertex(this.E);
		if ( this.r[this.E] >= this.r[this.B] ) {
			// Expansion worked.
			// E >= B   use simplex B..NE and go to step 3
			useVertex = this.E;
		} else {
			// Expansion failed.
			// E < B   use simplex B..NR and go to step 3
			useVertex = this.R;
		}
	} else {
		// We're moving in an undesireable direction.  Try contracting.
		// R < N
		if ( this.r[this.R] == Number.NEGATIVE_INFINITY ) {
			// Boundary Violation
			this.EvaluateVertex(this.Cr); 	// evaluate Cr
			if ( this.r[this.Cr] != Number.NEGATIVE_INFINITY ) {
				// in bounds!   use B..NCr and go to step 3
				useVertex = this.Cr;
			} else {
				// left with a Cw retraction.   use B..NCw and go to step 3
				this.EvaluateVertex(this.Cw);
				useVertex = this.Cw;
			}
		} else {
			// No boundary violation
			if ( this.r[this.R] >= this.r[this.W] ) {
				// R >= W   calculate and evaluate Cr.  use simplex B..NCr and go to step 3
				this.EvaluateVertex(this.Cr);
				useVertex = this.Cr;
			} else {
				// R < W   calculate and evaluate Cw.  use simplex B..NCw and go to step 3
				this.EvaluateVertex(this.Cw);
				useVertex = this.Cw;
			}
		}
	}
	
	// step 3
	// discard W
	try {
		for ( d=0; d<this.nDim; d++ ) {	// for each dimension...
			this.v[this.W][d] = this.v[useVertex][d];
		}
	} catch (e) {
		alert(e + "\n useVertex=" + useVertex);
		alert("R=" + this.R);
		alert("E=" + this.E);
		alert("Cr=" + this.Cr);
		alert("Cw=" + this.Cw);
	}
	this.r[this.W] = this.r[useVertex];
	// transfer N to new W slot
	this.W = this.N;
	
	this.nSteps++;
	
	var ch, i, cacheIndex;
	for ( i=0; i<this.nDim + 1; i++ ) {
		ch = "   ";
		switch (i) {
			case this.B: ch = "B  "; break;
			case this.W: ch = "W  "; break;
		}
		this.DenormalizeVertex(i);
		cacheIndex = "";
		for ( d=0; d<this.nDim; d++ ) {
			cacheIndex += this.uv[i][d] + " ";
		}
	}
}

// -------------------------------------------------------
// GetElapsedMinutes()
// -------------------------------------------------------
Simplex.prototype['GetElapsedMinutes'] = function () {
	var now = new Date();
	
	return Math.round( ((now - this.startTime) / 1000 / 60) * 100) / 100;
}

// -------------------------------------------------------
// IsGoalReached()
// -------------------------------------------------------
Simplex.prototype['IsGoalReached'] = function () {
	var bGoalReached = 0;
	
	if ( this.optType < 0 ) {
		bGoalReached = (Math.abs(this.bestResult) < this.goal) ? 1 : 0;    // TODO: fix for negative goals
	} else if ( this.optType > 0 ) {
		bGoalReached = (this.bestResult > this.goal) ? 1 : 0;
	} else {
		bGoalReached = (Math.abs(this.bestResult) < Math.abs(this.goal)) ? 1 : 0;
	}
	
	return bGoalReached;
}

// -------------------------------------------------------
// IsDegenerate()
// Determines if simplex had degenerated to a single point
// or is oscillating
// -------------------------------------------------------
Simplex.prototype['IsDegenerate'] = function () {
	var i, d;
	var cacheIndex;
	var degenerate = 0;
	
	for ( i=0; i<this.nDim+1; i++ ) {
		this.DenormalizeVertex[i];
		cacheIndex = "";
		for ( d=0; d<this.nDim; d++ ) {
			cacheIndex += this.uv[i][d] + " ";
		}
		
		if ( this.evalCount[cacheIndex] > this.nDim*10 ) {
			degenerate = 1;
		}
	}
	
	return degenerate;
}

// -------------------------------------------------------
// Optimize()
// Begins optimization process.  Repeatedly calls Step()
// to do the optimization until one of the exit criteria
// defined by DefineOptimization() is reached.
// startingPoint must be an array of nDim user-space
// coordinates
// -------------------------------------------------------
Simplex.prototype['Optimize'] = function (startingPoint) {
	var d;
	var reason;
	
	display.DisplayStatus("");
	
	this.startTime = new Date();
	this.nSteps = 0;
	this.nRestarts = 0;
	this.bestResult = Number.NEGATIVE_INFINITY;
	
	// Evaluate starting point.
	for ( d=0; d<this.nDim; d++ ) {
		this.uv[0][d] = startingPoint[d];
		this.bestSettings[d] = startingPoint[d];
	}
	this.NormalizeVertex(0);
	this.EvaluateVertex(0);
	this.bestResult = this.r[0];
	
	// If goal reached on starting point, exit immediately.
	if ( this.IsGoalReached()  ) {
		reason = "Finished (Goal reached).";
	} else {
		
		try {
			this.Initialize(startingPoint);
		} catch (e) {
			alert(e);
		}
		
		// Optimize
		while ( 1 ) {
			if ( this.IsDegenerate() && this.B >= 0 ) {
				// simplex has degenerated to a point or is oscillating
				// restart using random points (best point is still retained)
				this.nRestarts++;
				
				if ( this.nRestarts <= this.restartLimit ) {
					for ( d=0; d<this.nDim; d++ ) {
						this.v[0][d] = Math.random();
					}
					this.DenormalizeVertex(0);
					this.Initialize(this.uv[0]);
				}
			}
			
			if ( this.IsGoalReached() && this.nRestarts > 0 ) {
				// goal reached and simplex converged at least once
				reason = "Finished (Goal reached).";
				break;
			}
			
			if ( this.GetElapsedMinutes() > this.timeLimit ) {
				// ran too long
				reason = "Terminated (Ran too long).";
				break;
			}
			
			if ( this.nSteps > this.stepLimit ) {
				// too many steps
				reason = "Terminated (Too many steps).";
				break;
			}
			
			if ( this.nRestarts > this.restartLimit ) {
				// too many restarts
				reason = "Terminated (Too many restarts).";
				break;
			}
			
			this.Step();
		}
	}
	
	for ( d=0; d<this.nDim; d++ ) {
		acr[this.fnSet[d]](this.bestSettings[d])
	}
	
	return this.bestSettings;
}


// =======================================================
// Display Routines
// =======================================================


// -------------------------------------------------------
// Display()
// Constructor
// -------------------------------------------------------
function Display() {
	
	this.files = new Array();			// array of debug output files
	
	this.docName = activeDocument.name;	// save document name so it can be selected later
	try {
		this.statusWnd = documents.getByName("ACR Calibrator Status Window");
	} catch (e) {
		// Doesn't exist
	}
	
	this.status = "";
	this.statusSpace = "";
	
	this.startTime = new Date();
	
	this.linesOfBlackText = 0;
	this.linesOfRedText = 0;
	this.linesOfGrayText = 0;
	
	this.refreshBlackText = false;
	this.refreshRedText = false;
	this.refreshGrayText = false;
	this.refreshResultsText = false;
	
	this.blackText = "";
	this.settingsText = "";
	this.redText = "";
	this.grayText = "";
	this.resultsText = "";
	
	// Mark all the settings with invalid data
	this.temp = Number.POSITIVE_INFINITY;
	this.tint = Number.POSITIVE_INFINITY;
	this.exps = Number.POSITIVE_INFINITY;
	this.shad = Number.POSITIVE_INFINITY;
	this.brgh = Number.POSITIVE_INFINITY;
	this.cntr = Number.POSITIVE_INFINITY;
	this.strt = Number.POSITIVE_INFINITY;
	this.blkb = Number.POSITIVE_INFINITY;
	this.rhue = Number.POSITIVE_INFINITY;
	this.rsat = Number.POSITIVE_INFINITY;
	this.ghue = Number.POSITIVE_INFINITY;
	this.gsat = Number.POSITIVE_INFINITY;
	this.bhue = Number.POSITIVE_INFINITY;
	this.bsat = Number.POSITIVE_INFINITY;
	
	this.NewStatusWindow();
}

// -------------------------------------------------------
// GetActiveDocPath()
// -------------------------------------------------------
Display.prototype['GetActiveDocPath'] = function () {
	return activeDocument.path.absoluteURI;
}

// -------------------------------------------------------
// GetActiveDocFullName()
// -------------------------------------------------------
Display.prototype['GetActiveDocFullName'] = function () {
	return activeDocument.fullName.absoluteURI;
}

// -------------------------------------------------------
// GetElapsedMinutes()
// -------------------------------------------------------
Display.prototype['GetElapsedMinutes'] = function () {
	var now = new Date();
	
	return Math.round( ((now - this.startTime) / 1000 / 60) * 100) / 100;
}

// -------------------------------------------------------
// SelectRawWnd()
// -------------------------------------------------------
Display.prototype['SelectRawWnd'] = function () {
	// Select raw file window
	try {
		activeDocument = documents.getByName(this.docName);
	} catch (e) {
		// Doesn't exist
	}
}

// -------------------------------------------------------
// SelectStatusWnd()
// -------------------------------------------------------
Display.prototype['SelectStatusWnd'] = function () {
	try {
		activeDocument = documents.getByName("ACR Calibrator Status Window");
	} catch (e) {
		// Doesn't exist
	}
}

// -------------------------------------------------------
// PurgeHistory()
// -------------------------------------------------------
Display.prototype['PurgeHistory'] = function () {
	var desc1 = new ActionDescriptor();
	desc1.putEnumerated( charIDToTypeID( "null" ), charIDToTypeID( "PrgI" ), charIDToTypeID( "Hsty" ) );
	executeAction( charIDToTypeID( "Prge" ), desc1, DialogModes.NO );
}

// -------------------------------------------------------
// UpdateSettingsText()
// -------------------------------------------------------
Display.prototype['UpdateSettingsText'] = function () {
	this.settingsText = "";
	
	if (this.temp != Number.POSITIVE_INFINITY)
		this.settingsText += "Temp. Color: " + this.temp + "\r";
	if (this.tint != Number.POSITIVE_INFINITY)
		this.settingsText += "Matiz: " + this.tint + "\r";
	if (this.exps != Number.POSITIVE_INFINITY)
		this.settingsText += "Exposicion: " + this.exps + "\r";
	if (this.shad != Number.POSITIVE_INFINITY)
		this.settingsText += "Negros: " + this.shad + "\r";
	if (this.brgh != Number.POSITIVE_INFINITY)
		this.settingsText += "Brillo: " + this.brgh + "\r";
	if (this.cntr != Number.POSITIVE_INFINITY)
		this.settingsText += "Contraste: " + this.cntr + "\r";
	if (this.strt != Number.POSITIVE_INFINITY)
		this.settingsText += "Saturacion: " + this.strt + "\r";
	if (this.blkb != Number.POSITIVE_INFINITY)
		this.settingsText += "Sombras: " + this.blkb + "\r";
	if (this.rhue != Number.POSITIVE_INFINITY)
		this.settingsText += "Tono Rojo: " + this.rhue + "\r";
	if (this.rsat != Number.POSITIVE_INFINITY)
		this.settingsText += "Saturacion Rojo: " + this.rsat + "\r";
	if (this.ghue != Number.POSITIVE_INFINITY)
		this.settingsText += "Tono Verde: " + this.ghue + "\r";
	if (this.gsat != Number.POSITIVE_INFINITY)
		this.settingsText += "Saturacion Verde: " + this.gsat + "\r";
	if (this.bhue != Number.POSITIVE_INFINITY)
		this.settingsText += "Tono Azul: " + this.bhue + "\r";
	if (this.bsat != Number.POSITIVE_INFINITY)
		this.settingsText += "Saturacion Azul: " + this.bsat + "\r";
}

// -------------------------------------------------------
// PostResults()
// -------------------------------------------------------
Display.prototype['PostResults'] = function (msg, bPass, score) {
	this.resultsText += msg + " " +  + Math.round(score*10000)/10000 + "\r";
	this.refreshResultsText = true;
	this.Refresh();
}


// -------------------------------------------------------
// Refresh()
// -------------------------------------------------------
Display.prototype['Refresh'] = function () {
	var i, space;
	var nTries = 0;
	var bSuccessful = false;
	
	while ( !bSuccessful && nTries < 3 ) {
		try {
			
			this.SelectStatusWnd();
			
			this.UpdateSettingsText();
			
			// count the lines
			this.linesOfBlackText = this.CountLines(this.blackText);
			this.linesOfBlackText += this.CountLines(this.settingsText);
			if ( this.refreshRedText ) {
				this.linesOfRedText = this.CountLines(this.redText);
			}
			if ( this.refreshGrayText ) {
				this.linesOfGrayText = this.CountLines(this.grayText);
			}
			
			// Display the text
			space = "";
			for ( i=0; i<(this.linesOfBlackText+this.linesOfRedText); i++ ) {
				space += "\r";
			}
			this.grayTextLayer.textItem.contents = space + this.grayText;
			space = "";
			for ( i=0; i<this.linesOfBlackText; i++ ) {
				space += "\r";
			}
			this.redTextLayer.textItem.contents = space + this.redText;
			this.blackTextLayer.textItem.contents = this.blackText + this.settingsText;
			
			if ( this.refreshResultsText )
				this.resultsTextLayer.textItem.contents = "\r\r\r\r\r\r" + this.resultsText;
			
			this.refreshBlackText = false;
			this.refreshRedText = false;
			this.refreshGrayText = false;
			this.refreshResultsText = false;
			
			this.WaitForRedraw();
			
			this.SelectRawWnd();
			
			bSuccessful = true;
		} catch (e) {
			// Status window not selected for some reason
			nTries++;
		}
	}
	
	if ( nTries >= 3 ) {
		throw ("Error writing to status window.  Script aborted.");
	}
	
	this.PurgeHistory();
}

// -------------------------------------------------------
// WaitForRedraw()
// -------------------------------------------------------
Display.prototype['WaitForRedraw'] = function () {
	var eventWait = charIDToTypeID('Wait');
	var enumRedrawComplete = charIDToTypeID('RdCm');
	var typeState = charIDToTypeID('Stte');
	var keyState = charIDToTypeID('Stte');
	var desc = new ActionDescriptor();
	desc.putEnumerated(keyState, typeState, enumRedrawComplete);
	executeAction(eventWait, desc, DialogModes.NO);
}

// -------------------------------------------------------
// CountLines()
// -------------------------------------------------------
Display.prototype['CountLines'] = function (data) {
	var lines = data.split("\r");
	return lines.length - 1;
}


// -------------------------------------------------------
// NewStatusWindow()
// -------------------------------------------------------
Display.prototype['NewStatusWindow'] = function () {
	var statusWnd, bOldWndExists = true;
	try {
		statusWnd = documents.getByName("ACR Calibrator Status Window");
	} catch (e) {
		bOldWndExists = false;
	}
	if ( bOldWndExists ) {
		statusWnd.close(SaveOptions.DONOTSAVECHANGES);
	}
	
	// Create new document
    var desc1 = new ActionDescriptor();
	var desc2 = new ActionDescriptor();
	desc2.putString( charIDToTypeID( "Nm  " ), "ACR Calibrator Status Window" );
	desc2.putClass( charIDToTypeID( "Md  " ), charIDToTypeID( "RGBM" ) );
	desc2.putUnitDouble( charIDToTypeID( "Wdth" ), charIDToTypeID( "#Rlt" ), 640.000000 );
	desc2.putUnitDouble( charIDToTypeID( "Hght" ), charIDToTypeID( "#Rlt" ), 480.000000 );
	desc2.putUnitDouble( charIDToTypeID( "Rslt" ), charIDToTypeID( "#Rsl" ), 72.000000 );
	desc2.putDouble( stringIDToTypeID( "pixelScaleFactor" ), 1.000000 );
	desc2.putEnumerated( charIDToTypeID( "Fl  " ), charIDToTypeID( "Fl  " ), charIDToTypeID( "Wht " ) );
	desc2.putInteger( charIDToTypeID( "Dpth" ), 8 );
	desc2.putString( stringIDToTypeID( "profile" ), "ProPhoto RGB" );
    desc1.putObject( charIDToTypeID( "Nw  " ), charIDToTypeID( "Dcmn" ), desc2 );
	
	executeAction(charIDToTypeID("Mk  "), desc1, DialogModes.NO);
	
	this.statusWnd = documents.getByName("ACR Calibrator Status Window");
	
	// Select Status Window
	activeDocument = this.statusWnd;
	
	// Create the text layers
	var color = new SolidColor();
	color.rgb = new RGBColor();
	this.blackTextLayer = activeDocument.artLayers.add();
	this.blackTextLayer.kind = LayerKind.TEXT;
	this.blackTextLayer.name = "blackText";
	this.blackTextLayer.textItem.position = Array(activeDocument.width*0.005, activeDocument.height*0.03);
	this.blackTextLayer.textItem.size = 14;
	this.blackTextLayer.textItem.font = "Verdana";
	color.rgb.red = 0;
	color.rgb.green = 0;
	color.rgb.blue = 0;
	this.blackTextLayer.textItem.color = color;
	
	this.redTextLayer = activeDocument.artLayers.add();
	this.redTextLayer.kind = LayerKind.TEXT;
	this.redTextLayer.name = "redText";
	this.redTextLayer.textItem.position = Array(activeDocument.width*0.005, activeDocument.height*0.03);
	this.redTextLayer.textItem.size = 14;
	this.redTextLayer.textItem.font = "Verdana";
	color.rgb.red = 255;
	color.rgb.green = 0;
	color.rgb.blue = 0;
	this.redTextLayer.textItem.color = color;
	
	this.grayTextLayer = activeDocument.artLayers.add();
	this.grayTextLayer.kind = LayerKind.TEXT;
	this.grayTextLayer.name = "grayText";
	this.grayTextLayer.textItem.position = Array(activeDocument.width*0.005, activeDocument.height*0.03);
	this.grayTextLayer.textItem.size = 14;
	this.grayTextLayer.textItem.font = "Verdana";
	color.rgb.red = 128;
	color.rgb.green = 128;
	color.rgb.blue = 128;
	this.grayTextLayer.textItem.color = color;
	
	this.resultsTextLayer = activeDocument.artLayers.add();
	this.resultsTextLayer.kind = LayerKind.TEXT;
	this.resultsTextLayer.name = "resultsText";
	this.resultsTextLayer.textItem.position = Array(activeDocument.width*0.605, activeDocument.height*0.03);
	this.resultsTextLayer.textItem.size = 14;
	this.resultsTextLayer.textItem.font = "Verdana";
	color.rgb.red = 0;
	color.rgb.green = 0;
	color.rgb.blue = 0;
	this.resultsTextLayer.textItem.color = color;
	
	this.SelectRawWnd();
}


// -------------------------------------------------------
// DisplayStatus()
// -------------------------------------------------------
Display.prototype['DisplayStatus'] = function (data) {
	var lines = data.split("\r");
	
	if ( data.length > 0 ) {
		this.redText = lines[0] + "\r";
		this.refreshRedText = true;
	}
	
	this.grayText = "";
	for ( i=1; i<lines.length; i++ ) {
		this.grayText += lines[i] + "\r";
	}
	
	if ( data.length > 0 && lines[0] != "Done.") {
		this.grayText += "  Elapsed Time: " + this.GetElapsedMinutes() + " mins.  Speed: " + Math.round(acr.nOpens/this.GetElapsedMinutes()*100)/100;
	}
	this.refreshGrayText = true;
	
	this.Refresh();
}

// -------------------------------------------------------
// AddStatusLine()
// -------------------------------------------------------
Display.prototype['AddStatusLine'] = function (status) {
	this.blackText += status + "\r";
	this.refreshBlackText = true;
}


// =======================================================
// Color Utility Functions
// =======================================================

// -------------------------------------------------------
// ProPhotoRGBColor()
// Constructor
// -------------------------------------------------------
function ProPhotoRGBColor() {
	// ProPhotoRGB specification from icc file
	this.vR = new Vector1x3(0x0CC34/0x10000, 0x049BD/0x10000, 0x00000/0x10000); // red
	this.vG = new Vector1x3(0x0229C/0x10000, 0x0B63E/0x10000, 0x00000/0x10000); // green
	this.vB = new Vector1x3(0x00806/0x10000, 0x00006/0x10000, 0x0D32D/0x10000); // blue
	this.vW = new Vector1x3(0x0F6D6/0x10000, 0x10000/0x10000, 0x0D32C/0x10000); // white
	this.gamma = 461/256;
	
	// Compute RGB/XYZ Conversion Matrix
	var vS = (new Matrix3x3(this.vR, this.vG, this.vB)).Inverse().Multiply(this.vW);
	this.M = new Matrix3x3(this.vR.Multiply(vS.a11), this.vG.Multiply(vS.a12), this.vB.Multiply(vS.a13));
}

// -------------------------------------------------------
// SetRGB()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['SetRGB'] = function (r, g, b) {
	this.R = r;
	this.G = g;
	this.B = b;
	
	this.ComputeXYZ();
	this.Computexy();
	this.ComputeLab();
	this.ComputeLuv();
	this.ComputeCHS();
}

// -------------------------------------------------------
// ComputeXYZ()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['ComputeXYZ'] = function () {
	// Convert to linear, normalized rgb
	var R = Math.pow( (this.R/255), this.gamma);
	var G = Math.pow( (this.G/255), this.gamma);
	var B = Math.pow( (this.B/255), this.gamma);
	var RGB = new Vector1x3(R, G, B);
	
	// Convert to CIE XYZ
	var XYZ = this.M.Multiply(RGB);
	this.X = XYZ.a11;
	this.Y = XYZ.a12;
	this.Z = XYZ.a13;
}

// -------------------------------------------------------
// Computexy()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['Computexy'] = function () {
	// Compute x,y chromaticities
	this.x = this.X / (this.X+this.Y+this.Z);
	this.y = this.Y / (this.X+this.Y+this.Z);
}

// -------------------------------------------------------
// ComputeLab()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['ComputeLab'] = function () {
	// Compute CIE L*a*b*
	var E = 216/24389;
	var K = 24389/27;
	
	var x2 = this.X/this.vW.a11;
	var y2 = this.Y/this.vW.a12;
	var z2 = this.Z/this.vW.a13;
	
	if ( x2 > E )
		Fx = Math.pow(x2, 1/3);
	else
		Fx = (K*x2 + 16) / 116;
	
	if ( y2 > E )
		Fy = Math.pow(y2, 1/3);
	else
		Fy = (K*y2 + 16) / 116;
	
	if ( z2 > E )
		Fz = Math.pow(z2, 1/3);
	else
		Fz = (K*z2 + 16) / 116;
	
	if ( y2 > E )
		this.L = 116*Fy-16;
	else
		this.L = K*y2;
	this.a = 500*(Fx-Fy);
	this.b = 200*(Fy-Fz);
}

// -------------------------------------------------------
// ComputeLuv()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['ComputeLuv'] = function () {
	// Compute u' and v'
	this.up = 4*this.X / (this.X + 15*this.Y + 3*this.Z);
	this.vp = 9*this.Y / (this.X + 15*this.Y + 3*this.Z);
	
	// Compute ur' and vr' (u' and v' of reference white)
	var urp = 4*this.vW.a11 / (this.vW.a11 + 15*this.vW.a12 + 3*this.vW.a13);
	var vrp = 9*this.vW.a12 / (this.vW.a11 + 15*this.vW.a12 + 3*this.vW.a13);
	
	// Compute CIE L*u*v*
	this.u = 13 * this.L * (this.up - urp); // u*
	this.v = 13 * this.L * (this.vp - vrp); // v*
}

// -------------------------------------------------------
// ComputeCHS()
// -------------------------------------------------------
ProPhotoRGBColor.prototype['ComputeCHS'] = function () {
	this.C = Math.sqrt( (this.a*this.a + this.b*this.b), 2 );	// chroma
	this.H = Math.atan2(this.a, this.b) * 180 / Math.PI;		// hue
	this.S = this.C / this.L;									// saturation
}


// =======================================================
// 3x3 Matrix Routines
// =======================================================


// -------------------------------------------------------
// Matrix3x3()
// Constructor
// -------------------------------------------------------
function Matrix3x3(a11, a12, a13, a21, a22, a23, a31, a32, a33) {
	if ( (a11 instanceof Vector1x3) && (a12 instanceof Vector1x3) && (a13 instanceof Vector1x3) ) {
		// Three vectors have been specified.  Constuct matrix with each vector as a row
		this.a11 = a11.a11;
		this.a12 = a11.a12;
		this.a13 = a11.a13;
		this.a21 = a12.a11;
		this.a22 = a12.a12;
		this.a23 = a12.a13;
		this.a31 = a13.a11;
		this.a32 = a13.a12;
		this.a33 = a13.a13;
	} else {
		// assume individual matrix elements have been specified and contruct it accordingly
		this.a11 = a11;
		this.a12 = a12;
		this.a13 = a13;
		this.a21 = a21;
		this.a22 = a22;
		this.a23 = a23;
		this.a31 = a31;
		this.a32 = a32;
		this.a33 = a33;
	}
}

// -------------------------------------------------------
// SetBradford()
// Sets matrix to Bradford Chromatic Adapatation Matrix
// -------------------------------------------------------
Matrix3x3.prototype['SetBradford'] = function () {
	this.a11 =  0.8951;		this.a12 = -0.7502;		this.a13 =  0.0389;
	this.a21 =  0.2664;		this.a22 =  1.7135;		this.a23 = -0.0685;
	this.a31 = -0.1614;		this.a32 =  0.0368;		this.a33 =  1.0296;
}

// -------------------------------------------------------
// Inverse()
// Returns inverse of matrix.  Does not alter this matrix.
// -------------------------------------------------------
Matrix3x3.prototype['Inverse'] = function () {
	var det = this.Determinant();
	if ( det == 0 )
		throw("Attempted inverse of singular matrix.");
	
	var inv = new Matrix3x3( 	this.det2x2(this.a22,this.a23,this.a32,this.a33)/det,
								this.det2x2(this.a13,this.a12,this.a33,this.a32)/det,
								this.det2x2(this.a12,this.a13,this.a22,this.a23)/det,
								this.det2x2(this.a23,this.a21,this.a33,this.a31)/det,
								this.det2x2(this.a11,this.a13,this.a31,this.a33)/det,
								this.det2x2(this.a13,this.a11,this.a23,this.a21)/det,
								this.det2x2(this.a21,this.a22,this.a31,this.a32)/det,
								this.det2x2(this.a12,this.a11,this.a32,this.a31)/det,
								this.det2x2(this.a11,this.a12,this.a21,this.a22)/det  );
	return inv;
}

// -------------------------------------------------------
// Determinant()
// Returns determinant of matrix.
// -------------------------------------------------------
Matrix3x3.prototype['Determinant'] = function () {
	var det =  this.a11*this.a22*this.a33;
	    det -= this.a11*this.a23*this.a32;
	    det -= this.a12*this.a21*this.a33;
	    det += this.a12*this.a23*this.a31;
	    det += this.a13*this.a21*this.a32;
	    det -= this.a13*this.a22*this.a31;
				
	return det;
}


// -------------------------------------------------------
// det2x2()
// Utility function to compute determinant of 2x2 matrix
// -------------------------------------------------------
Matrix3x3.prototype['det2x2'] = function (a11, a12, a21, a22) {
	return ( (a11 * a22) - (a12 * a21) );
}

// -------------------------------------------------------
// Multiply()
// Multiplies lhs (left hand side) and matrix together.
// lhs can be another 3x3 matrix, a 1x3 vector, or a scalar.
// Returns appropriate result.  This matrix is not changed.
// -------------------------------------------------------
Matrix3x3.prototype['Multiply'] = function (lhs) {
	if ( lhs instanceof Matrix3x3 ) {
		// Matrix * Matrix
		return new Matrix3x3(  lhs.a11*this.a11 + lhs.a12*this.a21 + lhs.a13*this.a31,
							   lhs.a11*this.a12 + lhs.a12*this.a22 + lhs.a13*this.a32,
							   lhs.a11*this.a13 + lhs.a12*this.a23 + lhs.a13*this.a33,
							   lhs.a21*this.a11 + lhs.a22*this.a21 + lhs.a23*this.a31,
							   lhs.a21*this.a12 + lhs.a22*this.a22 + lhs.a23*this.a32,
							   lhs.a21*this.a13 + lhs.a22*this.a23 + lhs.a23*this.a33,
							   lhs.a31*this.a11 + lhs.a32*this.a21 + lhs.a33*this.a31,
							   lhs.a31*this.a12 + lhs.a32*this.a22 + lhs.a33*this.a32,
							   lhs.a31*this.a13 + lhs.a32*this.a23 + lhs.a33*this.a33   );
	}
	else if ( lhs instanceof Vector1x3 ) {
		// Vector * Matrix
		return new Vector1x3(  lhs.a11*this.a11 + lhs.a12*this.a21 + lhs.a13*this.a31,
							   lhs.a11*this.a12 + lhs.a12*this.a22 + lhs.a13*this.a32,
							   lhs.a11*this.a13 + lhs.a12*this.a23 + lhs.a13*this.a33   );
	}
	else {
		// Scalar * Matrix
		return new Matrix3x3(  lhs*this.a11, lhs*this.a12, lhs*this.a13,
							   lhs*this.a21, lhs*this.a22, lhs*this.a23,
							   lhs*this.a31, lhs*this.a32, lhs*this.a33   );
	}
}

// -------------------------------------------------------
// Display()
// Uses an alert box to display matrix for debugging purposes
// -------------------------------------------------------
Matrix3x3.prototype['Display'] = function () {
	var msg = "";
	msg += Math.round(this.a11*1000000)/1000000 + "\t";
	msg += Math.round(this.a12*1000000)/1000000 + "\t";
	msg += Math.round(this.a13*1000000)/1000000 + "\n";
	
	msg += Math.round(this.a21*1000000)/1000000 + "\t";
	msg += Math.round(this.a22*1000000)/1000000 + "\t";
	msg += Math.round(this.a23*1000000)/1000000 + "\n";
	
	msg += Math.round(this.a31*1000000)/1000000 + "\t";
	msg += Math.round(this.a32*1000000)/1000000 + "\t";
	msg += Math.round(this.a33*1000000)/1000000 + "\n";
	
	alert(msg);
}


// =======================================================
// 1x3 Matrix Routines
// =======================================================


// -------------------------------------------------------
// Vector1x3()
// Constructor
// -------------------------------------------------------
function Vector1x3(a11, a12, a13) {
	this.a11 = a11;
	this.a12 = a12;
	this.a13 = a13;
}

// -------------------------------------------------------
// Multiply()
// Multiplies lhs (left hand side) and vector together.
// lhs must be a scalar.
// Returns appropriate result.  This vector is not changed.
// -------------------------------------------------------
Vector1x3.prototype['Multiply'] = function (lhs) {
	// Scalar * Vector
	return new Vector1x3(  lhs*this.a11, lhs*this.a12, lhs*this.a13  );
}

// -------------------------------------------------------
// Display()
// Uses an alert box to display vector for debugging purposes
// -------------------------------------------------------
Vector1x3.prototype['Display'] = function () {
	var msg = "";
	msg += Math.round(this.a11*1000000)/1000000 + "\t";
	msg += Math.round(this.a12*1000000)/1000000 + "\t";
	msg += Math.round(this.a13*1000000)/1000000 + "\n";
	
	alert(msg);
}

// =======================================================
// Main Code
// =======================================================

// Save user's preferences
var rulerUnits = preferences.rulerUnits;
var nonLinearHistory = preferences.nonLinearHistory;
var numberOfHistoryStates = preferences.numberOfHistoryStates;

try {
	// Ask to change any required preferences
	var msg = ""
	if ( rulerUnits != Units.PIXELS )
		msg += "   --> Ruler Units must be set to Pixels.\n";
	if ( nonLinearHistory )
		msg += "   --> Nonlinear History must be disabled.\n";
	if ( numberOfHistoryStates < 2 )
		msg += "   --> History States must be set to 2 or more.\n";
	
	if ( msg != "" ) {
		if (confirm("ACR Calibrator needs to change your Photoshop preferences:\n\n" + msg + "\n" +
					"Select YES to allow ACR Calibrator to change your Photoshop preferences.\n" +
					"Present settings will be restored when the script exits.\n\n" +
					"Select NO to leave your preferences unchanged and exit ACR Calibrator.\n\n" +
					"Warning: if ACR Calibrator crashes unexpectedly, you will need to restore your\n" +
					"preferences manually."))
		{
			preferences.rulerUnits = Units.PIXELS;
			preferences.nonLinearHistory = false;
			if ( numberOfHistoryStates < 2 )
				preferences.numberOfHistoryStates = 2;
		} else {
			throw ("User aborted.");
		}
	}
	
	// Initialize Display Module
	var display = new Display();
	
	// Initialize ACR
	var acr = new Acr();
	acr.Initialize();
	
	// Initialize ColorChecker
	var cc = new ColorChecker();
	cc.FindPatchesFromActiveDoc();
	
	display.AddStatusLine("AcrCalibrator - Version " + scriptVersion);
	display.AddStatusLine("Copyright (c) 2007 Thomas Fors <tom@fors.net>");
    display.AddStatusLine("Optimizado para tarjeta de calibracin www.copiasxl.com V1.05");
	 display.AddStatusLine("Modificado por CopiasXL para soportar las versiones ACR4.2 , ACR 4.3");
	display.AddStatusLine("--------------------------------------------------------");
	display.AddStatusLine("ACR Version " + acr.version + acr.supported);
	display.AddStatusLine(display.GetActiveDocFullName());
	display.AddStatusLine("Corners: " + cc.patchCornerString);
	display.AddStatusLine("");
	display.Refresh();
	
	// Close open file after gathering the useful info from it
	acr.Close();
	
	// Perform calibration
	cc.WhiteBalance(1)
	cc.GrayAdjust();
	cc.WhiteBalance(2);
	cc.SaturationAdjust();
	cc.ShadowTintAdjust();
	cc.GreenCalibration(1);
	cc.BlueCalibration(1);
	cc.RedCalibration(1);
	cc.GreenCalibration(2);
	cc.BlueCalibration(2);
	cc.RedCalibration(2);
	cc.GreenCalibration(3);
	cc.BlueCalibration(3);
	cc.RedCalibration(3);
	
	display.DisplayStatus("Done.\rElapsed Time: " + display.GetElapsedMinutes() + " mins.  Speed: " + Math.round(acr.nOpens/display.GetElapsedMinutes()*100)/100);
	
	// Restore user's ruler units
	preferences.rulerUnits = rulerUnits;
	preferences.nonLinearHistory = nonLinearHistory;
	preferences.numberOfHistoryStates = numberOfHistoryStates;
	
} catch (e) {
	
	// Restore user's ruler units
	preferences.rulerUnits = rulerUnits;
	preferences.nonLinearHistory = nonLinearHistory;
	preferences.numberOfHistoryStates = numberOfHistoryStates;
	
	alert(e);
}
